from typing import Optional, List

from pydantic import BaseModel


class PlaywrightAction(BaseModel):
    """
    The BaseModel for playwright automations

    Goal:
        This contains an exhaustive list of commands that playwright can execute. It
        will be filled accordingly by the LLM depending on the DOM recieved from playwright
        and the goal of the task.
    """

    goto: Optional[str]
    go_back: Optional[bool]
    go_forward: Optional[bool]
    reload: Optional[bool]

    click: Optional[str]
    dblclick: Optional[str]
    hover: Optional[str]
    fill_selector: Optional[str]
    fill_value: Optional[str]
    type_selector: Optional[str]
    type_text: Optional[str]
    press_selector: Optional[str]
    press_key: Optional[str]
    check: Optional[str]
    uncheck: Optional[str]
    select_selector: Optional[str]
    select_value: Optional[str]
    upload_selector: Optional[str]
    upload_path: Optional[str]

    scroll_x: Optional[int]
    scroll_y: Optional[int]
    wait_selector: Optional[str]
    wait_timeout: Optional[int]
    wait_ms: Optional[int]

    keyboard_press: Optional[str]
    keyboard_type: Optional[str]
    mouse_move_x: Optional[int]
    mouse_move_y: Optional[int]
    mouse_click_x: Optional[int]
    mouse_click_y: Optional[int]

    new_page: Optional[str]
    close_page: Optional[bool]
    switch_page_index: Optional[int]

    evaluate_js: Optional[str]
    screenshot_path: Optional[str]
    download_selector: Optional[str]


class PlaywrightResponse(BaseModel):
    actions: List[PlaywrightAction]



import math
from collections import Counter


def url_entropy(url):
    counts = Counter(url)
    total = len(url)
    return -sum((count / total) * math.log2(count / total) for count in counts.values())



import asyncio
import sys
from typing import List

from playwright.async_api import async_playwright

from pyba.core.agent.playwright_agent import PlaywrightAgent
from pyba.core.lib import DOMExtraction, HandleDependencies
from pyba.core.lib.action import perform_action
from pyba.core.scripts import LoginEngine
from pyba.utils.exceptions import (
    PromptNotPresent,
    ServiceNotSelected,
    ServerLocationUndefined,
    UnknownSiteChosen,
)


class Engine:
    """
    The main entrypoint for browser automation. This engine exposes the main entry point which is the run() method
    """

    def __init__(
        self,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        headless: bool = False,
        handle_dependencies: bool = True,
    ):
        """
        1. Get the keys
        2. Check if we are to run in the headless mode (default at No)
        3. Does the user need us to download dependencies (default at Yes)
        """
        self.provider = None
        self.headless_mode = headless
        self.automated_login_engine_classes = []

        # This needs to go inside a config file
        selectors = (
            "input:not([type='hidden']):not([type='submit']):not([type='button']):not([type='reset']):not([type='file'])",
            "textarea",
            "select",
            "[contenteditable='true']",
            "[role='textbox']",
            "[role='searchbox']",
            "[role='combobox']",
        )
        self.combined_selector = ", ".join(selectors)

        self.handle_dependencies(handle_dependencies)
        self.handle_keys(openai_api_key, vertexai_project_id, vertexai_server_location)
        # Defining the playwright agent with the defined configs
        self.playwright_agent = PlaywrightAgent(engine=self)  # This is amusing

    def handle_dependencies(self, handle_dependencies: bool):
        if handle_dependencies:
            HandleDependencies.playwright.handle_dependencies()

    def handle_keys(self, openai_api_key, vertexai_project_id, vertexai_server_location):
        if openai_api_key is None and vertexai_project_id is None:
            raise ServiceNotSelected()

        if vertexai_project_id and vertexai_server_location is None:
            raise ServerLocationUndefined(vertexai_server_location)

        if openai_api_key and vertexai_project_id:
            print(
                "You've defined both vertexai and openai models, we're choosing to go with openai!"
            )
            self.openai_api_key = openai_api_key
            self.provider = "openai"

        if vertexai_project_id:
            # Assuming that we don't have an openai_api_key
            self.provider = "vertexai"
            self.vertexai_project_id = vertexai_project_id
            self.location = vertexai_server_location
            self.model = "gemini-2.5-pro"  # Keeping this as fixed
        else:
            self.provider = "openai"
            self.openai_api_key = openai_api_key

    async def run(self, prompt: str = None, automated_login_sites: List[str] = None):
        """
        The most basic implementation for the run function

        Args:
            `prompt`: The user's instructions
                Right now we're assuming that the user's prompt is well defined. In later
                versions we'll come up with a fix for that as well.
        """

        if prompt is None:
            raise PromptNotPresent()

        if automated_login_sites is not None:
            assert isinstance(
                automated_login_sites, list
            ), "Make sure the automated_login_sites is a list!"

            for engine in automated_login_sites:
                # Each engine is going to be a name like "instagram"
                if hasattr(LoginEngine, engine):
                    engine_class = getattr(LoginEngine, engine)
                    self.automated_login_engine_classes.append(engine_class)
                else:
                    raise UnknownSiteChosen(LoginEngine.available_engines())

        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=False)
            context = await browser.new_context()
            page = await context.new_page()

            cleaned_dom = {
                "hyperlinks": None,
                "input_fields": None,
                "clickable_fields": None,
                "actual_text": None,
                "current_url": None,
            }

            for steps in range(0, 10):
                # First check if we need to login and run the scripts
                # If loginengines have been chosen then self.automated_login_engine_classes will be populated
                if self.automated_login_engine_classes:
                    for engine in self.automated_login_engine_classes:
                        engine_instance = engine(page)
                        # Instead of just running it and checking inside, we can have a simple lookup list
                        out_flag = await engine_instance.run()
                        if out_flag:
                            # This means it was True and we successfully logged in
                            print(f"Logged in successfully through the {page.url} link")
                        elif out_flag is None:
                            # This means it wasn't for a login page for this engine
                            print("skipping")
                            pass
                        else:
                            # This means it failed
                            print(f"Login attempted at {page.url} but failed!")

                # Say we're going to run only 10 steps so far, so after this no more automation
                # Get an actionable PlaywrightResponse from the models
                action = self.playwright_agent.process_action(
                    cleaned_dom=cleaned_dom, user_prompt=prompt
                )

                if all(value is None for value in vars(action).values()):
                    # This means the goal has been achieved
                    print("Automated completed, agent returned None")
                    sys.exit(0)

                print(f"\n\nThis is the action: {action}\n\n")
                # If its not None, then perform it
                await perform_action(page, action)

                page_html = await page.content()
                body_text = await page.inner_text("body")
                elements = await page.query_selector_all(self.combined_selector)
                base_url = page.url

                # Then we need to extract the new cleaned_dom from the page
                extractionEngine = DOMExtraction(
                    html=page_html, body_text=body_text, elements=elements, base_url=base_url
                )

                # Passing in known_fields for the input fields that we already know off so that
                # its easier for the extraction engine to work
                cleaned_dom = await extractionEngine.extract()

                cleaned_dom["current_url"] = base_url

    def sync_run(self, prompt: str = None, automated_login_sites: List[str] = None):
        """
        Sync endpoint for running the above function
        """
        asyncio.run(self.run(prompt=prompt, automated_login_sites=automated_login_sites))



import subprocess
import sys


class PlaywrightDependencies:
    """
    We'll make use of this class if the user wants us to handle the dependencies. Usually, this
    shouldn't be the case because its like 2 commands to manage all playwright deps.
    """

    @staticmethod
    def check_playwright_browsers_installed() -> bool:
        """
        Check if Playwright browsers are installed by running `playwright install --dry-run`
        """
        try:
            result = subprocess.run(
                [sys.executable, "-m", "playwright", "install", "--dry-run"],
                capture_output=True,
                text=True,
                check=False,
            )
            # If dry-run shows no browsers to install, they are already installed
            return "No browsers to install" in result.stdout
        except Exception:
            return False

    @staticmethod
    def install_playwright_browsers():
        """
        Install Playwright browsers automatically.
        """
        print("Installing Playwright browsers...")
        subprocess.run([sys.executable, "-m", "playwright", "install"], check=True)

    @staticmethod
    def check_missing_dependencies():
        """
        Run a test to identify missing dependencies using `playwright install-deps --dry-run`.
        """
        try:
            result = subprocess.run(
                [sys.executable, "-m", "playwright", "install-deps", "--dry-run"],
                capture_output=True,
                text=True,
                check=False,
            )
            output = result.stdout.strip() + "\n" + result.stderr.strip()
            # Filter only the lines that mention missing libraries
            missing_lines = [
                line
                for line in output.splitlines()
                if "Missing" in line or line.strip().endswith(".so") or "║" in line
            ]
            if missing_lines:
                print("The following dependencies are missing: ")
                print("\n".join(missing_lines))
                print(
                    "Please install them manually using your package manager or run: playwright install-deps"
                )

        except FileNotFoundError:
            print("Could not run playwright. Is it installed in this environment?")

    @staticmethod
    def handle_dependencies():
        # Step 1: Check if browsers are installed
        if PlaywrightDependencies.check_playwright_browsers_installed():
            print("Playwright browsers are already installed")
        else:
            print("Playwright browsers not found.")
            choice = input("Do you want to install them automatically? (y/n): ").strip().lower()
            if choice == "y":
                PlaywrightDependencies.install_playwright_browsers()
            else:
                print("Please install browsers using: playwright install")
        # Step 2: Check missing system dependencies
        PlaywrightDependencies.check_missing_dependencies()


class HandleDependencies:
    playwright = PlaywrightDependencies



import json
from types import SimpleNamespace
from typing import Dict, List, Union

from google import genai
from google.genai.types import GenerateContentConfig
from openai import OpenAI

from pyba.utils.prompts import system_instruction, general_prompt
from pyba.utils.structure import PlaywrightResponse


class PlaywrightAgent:
    """
    The automation main engine. This takes in all the context from
    the current screen and decides the next best action. We're
    supporting two APIs for now:

    1. VertexAI projects
    2. OpenAI

    Depending on which key you have set in the main engine, this agent
    will be called. This configuration is taken directly from the engine.
    We're not making this an inherited class of Engine because this is
    technically not an Engine per se, its is own thing.
    """

    def __init__(self, engine) -> None:
        """
        `engine` basically holds all the arguments from the
        """
        self.engine = engine
        self.initialize_playwright_agent()

    def initialize_playwright_agent(self) -> None:
        """
        Initialises a client/agent depending on the provider
        """
        if self.engine.provider == "openai":
            self.openai_client = OpenAI(api_key=self.engine.openai_api_key)
            self.agent = {
                "client": self.openai_client,
                "system_instruction": system_instruction,
                "model": "gpt-4o",
                "response_format": PlaywrightResponse,
            }
        else:
            self.vertexai_client = genai.Client(
                vertexai=True,
                project=self.engine.vertexai_project_id,
                location=self.engine.location,
            )

            self.agent = self.vertexai_client.chats.create(
                model=self.engine.model,
                config=GenerateContentConfig(
                    temperature=0,
                    system_instruction=system_instruction,
                    response_schema=PlaywrightResponse,
                    response_mime_type="application/json",
                ),
            )

    def process_action(
        self, cleaned_dom: Dict[str, Union[List, str]], user_prompt: str
    ) -> PlaywrightResponse:
        """
        Method to process the DOM and provide an actionable playwright response

        Args:
            `cleaned_dom`: Dictionary of the extracted items from the DOM
                - `hyperlinks`: List
                - `input_fields` (basically all fillable boxes): List
                - `clickable_fields`: List
                - `actual_text`: string
            `user_prompt`: The instructions given by the user

            We're assuming this to be well explained. In later versions we'll
            add one more layer on top for plan generation and better commands

            output: A predefined pydantic model
        """
        cleaned_dom["user_prompt"] = user_prompt
        prompt = general_prompt.format(**cleaned_dom)

        print()
        print(cleaned_dom)
        print()

        if self.engine.provider == "openai":
            messages = [
                {"role": "system", "content": self.agent["system_instruction"]},
                {"role": "user", "content": prompt},
            ]
            kwargs = {
                "model": self.agent["model"],
                "messages": messages,
            }
            # This is when we are passing a response scehma to the model. We use the .parse() endpoint
            response = self.agent["client"].chat.completions.parse(
                **kwargs, response_format=self.agent["response_format"]
            )
            return SimpleNamespace(
                **json.loads(response.choices[0].message.content).get("actions")[0]
            )
        else:
            response = self.agent.send_message(prompt)
            try:
                # We should prefer .output_parsed if using google-genai structured output
                actions = getattr(response, "output_parsed", getattr(response, "parsed", None))
                if actions and hasattr(actions, "actions") and actions.actions:
                    return actions.actions[0]
                raise IndexError("No actions found in response.")
            except Exception as e:
                print(f"Unable to parse the output from VertexAI response: {e}")



import os
from typing import Optional
from urllib.parse import urlparse

from dotenv import load_dotenv
from playwright.async_api import Page

from pyba.utils.exceptions import CredentialsnotSpecified

load_dotenv()  # Loading the username and passwords

SCREEN_HEIGHT = 1280  # my screen's height, we need this to be in the config as well!
x_from_left = 1000
y_from_bottom = 1000
y_top_left = SCREEN_HEIGHT - y_from_bottom


class InstagramLogin:
    """
    The instagram login engine
    """

    def __init__(self, page: Page) -> None:
        self.page = (
            page  # This is the page we're at, this is where the login automation needs to happen
        )

        self.engine_name = "instagram"
        self.username = os.getenv("instagram_username")
        self.password = os.getenv("instagram_password")

        if self.username is None or self.password is None:
            raise CredentialsnotSpecified(self.engine_name)

        self.uses_2FA = False

    def verify_login_page(self):
        """
        Make sure that the script we're going to run is made for this login page itself. This uses multiple ways to
        ensure that. First it verifies it through the URL, then checks if the elements we are going to append to.
        """

        page_url = self.page.url

        # This needs to be taken from a config file. I will start maintaing that later
        instagram_urls = [
            "https://www.instagram.com/accounts/login/",
            "https://www.instagram.com/",
        ]

        # We'll have to clean the URL from all the url formatting to the basic thing and match it with this.
        # This can be done using urlparse and normalizing it first
        parsed = urlparse(page_url)
        normalized_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"

        if not normalized_url.endswith("/"):
            normalized_url += "/"

        # Keeping it simple with this right now, later we can make this better
        if normalized_url in instagram_urls:
            return True
        else:
            return False

    async def run(self) -> Optional[bool]:
        """
        The idea is to take in the username and password from the .env file for now
        and simply use that to execute this function

        Returns:
                `None` if we're not supposed to launch the automated login script here
                `True/False` if the login was successful or a failure
        """
        val = self.verify_login_page()
        if not val:
            return None

        # Now run the script
        try:
            await self.page.wait_for_selector('input[name="username"]')
            await self.page.fill('input[name="username"]', self.username)
            await self.page.fill('input[name="password"]', self.password)

            await self.page.click('button[type="submit"]')
        except Exception:
            # Now this is bad
            return False

        # There is a not-now button that we need to click for not saving our information
        try:
            await self.page.wait_for_selector('text="Not now"', timeout=30000)
            await self.page.click('text="Not now"')
        except Exception:
            # This means that never came so we're done.
            pass

        # Sometimes these things also come up for new updates
        try:
            await self.page.wait_for_selector('text="OK"', timeout=10000)
            await self.page.mouse.click(x_from_left, y_top_left)
        except Exception:
            # This means this never came up
            pass

        try:
            await self.page.wait_for_load_state("networkidle", timeout=10000)
        except Exception:
            # It's fine, we'll assume that the login worked nicely
            pass

        return True



from playwright.async_api import Page

from pyba.utils.structure import PlaywrightAction


class PlaywrightActionPerformer:
    """
    The playwright automation class. To add new handles, make a function here
    and define that under perform()
    """

    def __init__(self, page: Page, action: PlaywrightAction):
        self.page = page
        self.action = action

    async def handle_navigation(self):
        """
        Handle's browser naviation -> Opening new websites
        Wait's until the page is loaded
        """
        await self.page.goto(self.action.goto)
        await self.page.wait_for_load_state("domcontentloaded")

    async def handle_input(self):
        """
        Inputs a value to a selector field
        """
        await self.page.fill(self.action.fill_selector, self.action.fill_value)

    async def handle_click(self):
        """
        Handle's clicking elements. Has additional checks to ensure that
        the element is not actually a relational hyperlink
        """
        click_target = self.action.click

        # Handle relational hyperlinks seperately
        if click_target.startswith("a[href=") or "href=" in click_target:
            href = click_target.split('"')[1]
            if href.startswith("/"):
                base_url = self.page.url.split("/")[0:3]
                href = "/".join(base_url) + href
            await self.page.goto(href)
            await self.page.wait_for_load_state("domcontentloaded")
        else:
            locator = self.page.locator(click_target)
            try:
                await locator.scroll_into_view_if_needed()
                await locator.click(timeout=5000)
            except Exception:
                # Try to force a click
                await locator.click(force=True, timeout=5000)

    async def handle_double_click(self):
        """
        Handle's double clicking an element
        """
        await self.page.dblclick(self.action.dblclick)

    async def handle_hover(self):
        """
        Handle's hovering over an element to make new actions visible
        """
        await self.page.hover(self.action.hover)

    async def handle_press(self):
        """
        Handles a key press.
        """
        # If a specific selector is provided, press the key on that element.
        if self.action.press_selector and self.action.press_key:
            await self.page.press(self.action.press_selector, self.action.press_key)
        # If no selector is provided, press the key on the entire page.
        elif self.action.press_key:
            await self.page.keyboard.press(self.action.press_key)

    async def handle_keyboard_press(self):
        """
        Handles a keyboard press action on the entire page.
        """
        await self.page.keyboard.press(self.action.keyboard_press)

    async def handle_checkboxes(self):
        """
        Checking and unchecking of boxes
        """
        if self.action.check:
            await self.page.check(self.action.check)
        if self.action.uncheck:
            await self.page.uncheck(self.action.uncheck)

    async def handle_scrolling(self):
        """
        Automates manual scrolling (or scrolls to center)
        """
        x = self.action.scroll_x or 0
        y = self.action.scroll_y or 0
        await self.page.mouse.wheel(x, y)

    async def perform(self) -> None:
        """
        This is the main dispatch function. All handlers are called here
        as and when required by the AI models.
        """
        action = self.action

        if action.goto:
            await self.handle_navigation()
            return

        if action.fill_selector and action.fill_value is not None:
            await self.handle_input()
            return

        if action.click:
            await self.handle_click()
            return

        if action.dblclick:
            await self.handle_double_click()
            return

        if action.hover:
            await self.handle_hover()
            return

        # Handle the specific case of press_selector and press_key together
        if action.press_selector and action.press_key:
            await self.handle_press()
            return

        # Handle the specific case of keyboard_press on its own
        if action.keyboard_press:
            await self.handle_keyboard_press()
            return

        if action.check or action.uncheck:
            await self.handle_checkboxes()
            return

        if action.scroll_x or action.scroll_y:
            await self.handle_scrolling()
            return


async def perform_action(page: Page, action: PlaywrightAction) -> None:
    """
    The entry point function
    """
    # assert isinstance(action, PlaywrightAction), "the input type for action is incorrect!"
    performer = PlaywrightActionPerformer(page, action)
    await performer.perform()



from pyba.core.scripts.login import InstagramLogin


class LoginEngine:
    instagram = InstagramLogin

    @classmethod
    def available_engines(cls):
        return [name for name, value in vars(cls).items() if isinstance(value, type)]
